class SPSSConverterError(ValueError):
    """Base exception raised by the **SPSS Converter** library."""
    pass


class ColumnNameNotFoundError(ValueError):
    """Exception raised when a given column or variable name is not found."""
    pass


class InvalidDataFormatError(ValueError):
    """Exception raised when a value did not conform to an expected type."""
    pass


class InvalidLayoutError(ValueError):
    """Exception raised when a layout value was not recognized."""
    pass



from enum import Enum
from typing import Union

from validator_collection import validators, checkers
from spss_converter import errors
from pyreadstat._readstat_parser import metadata_container


class VariableAlignmentEnum(str, Enum):
    UNKNOWN = 'unknown'
    LEFT = 'left'
    CENTER = 'center'
    RIGHT = 'right'


class VariableMeasureEnum(str, Enum):
    UNKNOWN = 'unknown'
    NOMINAL = 'nominal'
    ORDINAL = 'ordinal'
    SCALE = 'scale'


class ColumnMetadata(object):
    """Object representation of the :term:`metadata <Metadata>` that describes a column or
    variable form an SPSS file."""

    def __init__(self, **kwargs):
        self._name = None
        self._label = None
        self._value_metadata = None
        self._missing_range_metadata = None
        self._missing_value_metadata = None
        self._alignment = VariableAlignmentEnum.UNKNOWN
        self._measure = VariableMeasureEnum.UNKNOWN
        self._display_width = None
        self._storage_width = None

        for key in kwargs:
            value = kwargs.get(key)
            setattr(self, key, value)

    @property
    def name(self):
        """The name of the column/variable.

        :rtype: :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = validators.variable_name(value, allow_empty = True)

    @property
    def label(self):
        """The label applied ot the column/variable.

        :rtype: :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    def label(self, value):
        self._label = validators.string(value, allow_empty = True)

    @property
    def alignment(self):
        """The alignment to apply to values from this column/variable when displaying
        data. Defaults to ``'unknown'``.

        Accepts either ``'unknown'``, ``'left'``, ``'center'``, or ``'right'`` as
        either a case-insensitive :class:`str <python:str>` or a
        :class:`VariableAlignmentEnum`.

        :rtype: :class:`VariableAlignmentEnum`
        """
        return self._alignment

    @alignment.setter
    def alignment(self, value):
        value = validators.string(value, allow_empty = False)
        value = value.lower()
        if value not in [member.value
                         for name, member in VariableAlignmentEnum.__members__.items()]:
            raise ValueError(f'value ({value}) is not a recognized alignment')

        self._alignment = value

    @property
    def measure(self):
        """A classification of the type of measure (or value type) represented by the
        variable. Defaults to ``'unknown'``.

        Accepts either ``'unknown'``, ``'nominal'``, ``'ordinal'``, or ``'scale'``.

        :rtype: :class:`VariableMeasureEnum`
        """
        return self._measure

    @measure.setter
    def measure(self, value):
        value = validators.string(value, allow_empty = False)
        value = value.lower()
        if value not in [member.value
                         for name, member in VariableMeasureEnum.__members__.items()]:
            raise ValueError(f'value ({value}) is not a recognized measure')

        self._measure = value

    @property
    def display_width(self):
        """The maximum width at which the value is displayed. Defaults to 0.

        :rtype: :class:`int <python:int>`
        """
        return self._display_width

    @display_width.setter
    def display_width(self, value):
        self._display_width = validators.integer(value,
                                                 allow_empty = False,
                                                 minimum = 0,
                                                 coerce_value = True)

    @property
    def storage_width(self):
        """The width of data to store in the data file for the value. Defaults to 0.

        :rytpe: :class:`int <python:int>`
        """
        return self._storage_width

    @storage_width.setter
    def storage_width(self, value):
        self._storage_width = validators.integer(value,
                                                 allow_empty = False,
                                                 minimum = 0,
                                                 coerce_value = True)

    @property
    def value_metadata(self):
        """Collection of values possible for the column/variable, with corresponding
        labels for each value.

        :returns: :class:`dict <python:dict>` whose keys are the values in the raw data
          and whose values are the labels for each value. May be :obj:`None <python:None>`
          for variables whose value is not coded.
        :rtype: :class:`dict <python:dict>` / :obj:`None <python:None>`
        """
        return self._value_metadata

    @value_metadata.setter
    def value_metadata(self, value):
        value = validators.dict(value, allow_empty = True)
        if not value:
            self._value_metadata = None
        else:
            self._value_metadata = {
                key: validators.string(value[key], allow_empty = True) for key in value
            }

    @property
    def missing_range_metadata(self):
        """Collection of meta data that defines the numerical ranges that are to be
        considered missing in the underlying data.

        :returns: :class:`list <python:list>` of :class:`dict <python:dict>` with keys
          ``'low'`` and ``'high'`` for the low/high values of the range to apply when
          raw values are missing (:obj:`None <python:None>`).
        :rtype: :class:`list <python:list>` of :class:`dict <python:dict>` or
          :obj:`None <python:None>`

        """
        return self._missing_range_metadata

    @missing_range_metadata.setter
    def missing_range_metadata(self, value):
        value = validators.iterable(value, allow_empty = True)
        if not value:
            self._missing_range_metadata = None
        else:
            ranges = [validators.dict(x, allow_empty = False) for x in value]
            validated_ranges = []
            for range in ranges:
                if 'high' not in range or 'low' not in range:
                    raise ValueError('missing_range_metadata requires a "high" and "low"'
                                     ' boundary to be defined.')

                validated_range = {
                    'high': validators.numeric(range.get('high'), allow_empty = False),
                    'low': validators.numeric(range.get('low'), allow_empty = False)
                }

                validated_ranges.append(validated_range)

            self._missing_range_metadata = validated_ranges

    @property
    def missing_value_metadata(self):
        """Value used to represent misisng values in the raw data. Defaults to
        :obj:`None <python:None>`.

        .. note::

          This is not actually relevant for SPSS data, but is an artifact for SAS and SATA
          data.

        :rtype: :class:`list <python:list>` of :class:`int <python:int>` or
          :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._missing_value_metadata

    @missing_value_metadata.setter
    def missing_value_metadata(self, value):
        if not value:
            self._missing_value_metadata = None
            return
        elif checkers.is_string(value):
            value = [value]
        elif checkers.is_numeric(value):
            value = [value]

        validated_values = []
        for item in value:
            try:
                validated_item = validators.string(item, allow_empty = False)
            except (ValueError, TypeError):
                validated_item = validators.int(item, allow_empty = False)

            validated_values.append(validated_item)

        self._missing_value_metadata = validated_values

    @classmethod
    def from_dict(cls, as_dict: dict):
        """Create a new :class:`ColumnMetadata` instance from a
        :class:`dict <python:dict>` representation.

        :param as_dict: The :class:`dict <python:dict>` representation of the
          :class:`ColumnMetadata`.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The :class:`ColumnMetadata` instance.
        :rtype: :class:`ColumnMetadata`

        """
        instance = cls()
        instance.name = as_dict.pop('name', instance.name)
        instance.label = as_dict.pop('label', instance.label)
        instance.alignment = as_dict.pop('alignment', instance.alignment)
        instance.measure = as_dict.pop('measure', instance.measure)
        instance.display_width = as_dict.pop('display_width', instance.display_width)
        instance.storage_width = as_dict.pop('storage_width', instance.storage_width)
        instance.value_metadata = as_dict.pop('value_metadata', instance.value_metadata)
        instance.missing_range_metadata = as_dict.pop('missing_range_metadata',
                                                      instance.missing_range_metadata)
        instance.missing_value_metadata = as_dict.pop('missing_value_metadata',
                                                      instance.missing_value_metadata)

        return instance

    @classmethod
    def from_pyreadstat_metadata(cls,
                                 name: str,
                                 as_metadata):
        """Create a new :class:`ColumnMetadata` instance from a
        `Pyreadstat <https://github.com/Roche/pyreadstat/>`_ metadata object.

        :param name: The name of the variable for which a :class:`ColumnMetadata`
          instance should be created.
        :type name: :class:`str <python:str>`

        :param as_metadata: The `Pyreadstat <https://github.com/Roche/pyreadstat/>`_
          metadata object from which the column's metadata should be extracted.
        :type as_metadata: :class:`Pyreadstat.metadata_container <pyreadstat:_readstat_parser.metadata_container>`

        :returns: The :class:`ColumnMetadata` instance.
        :rtype: :class:`ColumnMetadata`

        """
        name = validators.variable_name(name, allow_empty = False)
        if name not in as_metadata.column_names:
            raise errors.ColumnNameNotFoundError(f'column name ({name}) not found in '
                                                 'as_metadata')
        instance = cls(name = name)
        instance.label = as_metadata.column_names_to_labels.get(name, None)
        instance.alignment = as_metadata.variable_alignment.get(name, instance.alignment)
        instance.measure = as_metadata.variable_measure.get(name, instance.measure)
        instance.display_width = as_metadata.variable_display_width.get(name,
                                                                        instance.display_width)
        instance.storage_width = as_metadata.variable_storage_width.get(name,
                                                                        instance.storage_width)
        instance.value_metadata = as_metadata.variable_value_labels.get(name,
                                                                        instance.value_metadata)

        missing_ranges = as_metadata.missing_ranges.get(name, [])
        instance.missing_range_metadata = [
            {
                'low': x.get('lo'),
                'high': x.get('hi')
            } for x in missing_ranges
        ]

        instance.missing_value_metadata = as_metadata.missing_user_values.get(name, None)

        return instance

    def to_dict(self) -> dict:
        """Generate a :class:`dict <python:dict>` representation of the instance.

        :rtype: :class:`dict <python:dict>`
        """
        return {
            'name': self.name,
            'label': self.label,
            'alignment': self.alignment,
            'measure': self.measure,
            'display_width': self.display_width,
            'storage_width': self.storage_width,
            'value_metadata': self.value_metadata,
            'missing_range_metadata': self.missing_range_metadata,
            'missing_value_metadata': self.missing_value_metadata
        }

    def add_to_pyreadstat(self, pyreadstat):
        """Update ``pyreadstat`` to include the metadata for this column/variable.

        :param pyreadstat: The `Pyreadstat <https://github.com/Roche/pyreadstat/>`_
          metadata object where the :class:`ColumnMetadata` data should be updated.
        :type pyreadstat: :class:`metadata_container <pyreadstat:_readstat_parser.metadata_container`

        :returns: The `Pyreadstat <https://github.com/Roche/pyreadstat/>`_ metadata.
        :rtype: :class:`metadata_container <pyreadstat:_readstat_parser.metadata_container`
        """
        pyreadstat.column_names_to_labels[self.name] = self.label
        pyreadstat.variable_alignment[self.name] = str(self.alignment)
        pyreadstat.variable_measure[self.name] = str(self.measure)
        pyreadstat.variable_display_width[self.name] = self.display_width
        pyreadstat.variable_storage_width[self.name] = self.storage_width
        if self.value_metadata is not None:
            pyreadstat.variable_value_labels[self.name] = self.value_metadata
            pyreadstat.value_labels[self.label] = self.value_metadata
        pyreadstat.variable_to_label[self.name] = self.label

        if self.missing_range_metadata:
            pyreadstat.missing_ranges[self.name] = [
                {
                    'lo': x.get('low'),
                    'hi': x.get('high')
                } for x in self.missing_range_metadata
            ]
        if self.missing_value_metadata:
            pyreadstat.missing_user_values[self.name] = [x for x in self.missing_value_metadata]

        if self.name not in pyreadstat.column_names:
            pyreadstat.column_names.append(self.name)
            pyreadstat.column_labels.append(self.label)
        else:
            index = pyreadstat.column_names.index(self.name)
            pyreadstat.column_names[index] = self.name
            pyreadstat.column_labels[index] = self.label

        return pyreadstat


class Metadata(object):
    """Object representation of :term:`metadata <Metadata>` retrieved from an SPSS file.
    """

    def __init__(self, **kwargs):
        self._column_metadata = None
        self._notes = None
        self._file_encoding = None
        self._rows = 0
        self._table_name = None
        self._file_label = None

        for key in kwargs:
            setattr(self, key, kwargs.get(key))

    @property
    def column_metadata(self):
        """Collection of metadata that describes each column or variable within the
        dataset.

        :returns: A :class:`dict <python:dict>` where the key is the name of the
          column/variable and the value is a :class:`ColumnMetadata` object or compatible
          :class:`dict <python:dict>`.
        :rtype: :class:`dict <python:dict>` / :obj:`None <python:None>`

        """
        return self._column_metadata

    @column_metadata.setter
    def column_metadata(self, value):
        value = validators.dict(value, allow_empty = True)
        if not value:
            self._column_metadata = None
        else:
            result = {}
            for key in value:
                key = validators.variable_name(key, allow_empty = False)
                if checkers.is_type(value[key], 'ColumnMetadata'):
                    result[key] = value[key]
                else:
                    result[key] = ColumnMetadata.from_dict(result[key])

            self._column_metadata = result

    @property
    def file_encoding(self) -> str:
        """The file encoding for the dataset.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._file_encoding

    @file_encoding.setter
    def file_encoding(self, value):
        self._file_encoding = validators.string(value, allow_empty = True)

    @property
    def notes(self) -> str:
        """Set of notes related to the file.

        :rtype: :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._notes

    @notes.setter
    def notes(self, value):
        if checkers.is_iterable(value):
            value = '\n'.join(value)
        self._notes = validators.string(value, allow_empty = True)

    @property
    def table_name(self) -> Union[str, None]:
        """The name of the data table.

        :rtype: :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._table_name

    @table_name.setter
    def table_name(self, value):
        self._table_name = validators.variable_name(value, allow_empty = True)

    @property
    def file_label(self) -> Union[str, None]:
        """The file label.

        .. note::

          This property is irrelevant for SPSS, but is relevant for SAS data.

        :rtype: :class:`str <python:str>` / :obj:`None <python:None>`
        """
        return self._file_label

    @file_label.setter
    def file_label(self, value):
        self._file_label = validators.string(value, allow_empty = True)

    @property
    def columns(self):
        """The number of columns/variables in the dataset.

        :rtype: :class:`int <python:int>`
        """
        if not self.column_metadata:
            return 0

        return len(self.column_metadata)

    @property
    def rows(self):
        """The number of cases or rows in the dataset.

        :rtype: :class:`int <python:int>`
        """
        return self._rows

    @rows.setter
    def rows(self, value):
        self._rows = validators.integer(value,
                                        allow_empty = False,
                                        coerce_value = False,
                                        minimum = 0)

    @classmethod
    def from_dict(cls, as_dict: dict):
        """Create a :class:`Metadata` instance from a :class:`dict <python:dict>`
        representation.

        :param as_dict: A :class:`dict <python:dict>` representation of the
          :class:`Metadata`.
        :type as_dict: :class:`dict <python:dict>`

        :returns: A :class:`Metadata` instance
        :rtype: :class:`Metadata`

        """
        instance = cls()
        instance.notes = as_dict.get('notes', instance.notes)
        instance.table_name = as_dict.get('table_name', instance.table_name)
        instance.file_label = as_dict.get('file_label', instance.file_label)
        instance.rows = as_dict.get('rows', instance.rows)
        instance.file_encoding = as_dict.get('file_encoding', instance.file_encoding)
        instance.column_metadata = as_dict.get('column_metadata',
                                               instance.column_metadata)

        return instance

    def to_dict(self) -> dict:
        """Return a :class:`dict <python:dict>` representation of the instance.

        :rtype: :class:`dict <python:dict>`
        """
        return {
            'table_name': self.table_name,
            'file_label': self.file_label,
            'file_encoding': self.file_encoding,
            'columns': self.columns,
            'rows': self.rows,
            'column_metadata': self.column_metadata,
            'notes': self.notes
        }

    @classmethod
    def from_pyreadstat(cls, as_metadata):
        """Create a :class:`Metadata` instance from a
        `Pyreadstat <https://github.com/Roche/pyreadstat/>`_ metadata object.

        :param as_metadata: The `Pyreadstat <https://github.com/Roche/pyreadstat/>`_
          metadata object from which the :class:`Metadata` instance should be created.
        :type as_metadata: :class:`Pyreadstat.metadata_container <pyreadstat:_readstat_parser.metadata_container>`

        :returns: The :class:`Metadata` instance.
        :rtype: :class:`Metadata`
        """
        instance = cls()
        instance.notes = as_metadata.notes
        instance.table_name = as_metadata.table_name
        instance.file_label = as_metadata.file_label
        instance.rows = as_metadata.number_rows
        instance.file_encoding = as_metadata.file_encoding

        column_metadata = {}
        for x in as_metadata.column_names:
            column_metadata[x] = ColumnMetadata.from_pyreadstat_metadata(x, as_metadata)

        instance.column_metadata = {
            x: ColumnMetadata.from_pyreadstat_metadata(x, as_metadata)
            for x in as_metadata.column_names
        }

        return instance

    def to_pyreadstat(self):
        """Create a `Pyreadstat <https://github.com/Roche/pyreadstat/>`_ metadata
        representation of the :class:`Metadata` instance.

        :returns: The `Pyreadstat <https://github.com/Roche/pyreadstat/>`_ metadata.
        :rtype: :class:`metadata_container <pyreadstat:_readstat_parser.metadata_container`
        """
        as_metadata = metadata_container()
        as_metadata.table_name = self.table_name
        as_metadata.file_label = self.file_label
        as_metadata.file_encoding = self.file_encoding
        if self.notes and len(self.notes.split('\n')) > 1:
            notes = self.notes[0]
        else:
            notes = self.notes
        as_metadata.notes = notes
        as_metadata.rows = self.rows

        for column in self.column_metadata:
            as_metadata = self.column_metadata[column].add_to_pyreadstat(as_metadata)

        return as_metadata



from spss_converter.read import get_metadata, to_dataframe, to_csv, to_json, to_dict, \
    to_yaml, to_excel
from spss_converter.write import from_dataframe, from_csv, from_dict, from_yaml, \
    from_json, from_excel, apply_metadata
from spss_converter.Metadata import Metadata, ColumnMetadata


__all__ = [
    get_metadata,
    apply_metadata,
    to_dataframe,
    to_csv,
    to_dict,
    to_json,
    to_yaml,
    to_excel,
    from_dataframe,
    from_csv,
    from_dict,
    from_yaml,
    from_json,
    from_excel,
    Metadata,
    ColumnMetadata
]



from typing import Union, List, Optional, Any
import tempfile

import os
from io import BytesIO, StringIO
import simplejson as json
import yaml
import pyreadstat
from validator_collection import validators, checkers
from spss_converter import errors
from spss_converter.Metadata import Metadata
from pandas import ExcelWriter


def _read_spss(data: Union[bytes, BytesIO, 'os.PathLike[Any]'],
               limit: Optional[int] = None,
               offset: int = 0,
               exclude_variables: Optional[List[str]] = None,
               include_variables: Optional[List[str]] = None,
               metadata_only: bool = False,
               apply_labels: bool = False,
               labels_as_categories: bool = True,
               missing_as_NaN: bool = False,
               convert_datetimes: bool = True,
               dates_as_datetime64: bool = False,
               **kwargs):
    """Internal function that reads an SPSS (.sav or .zsav) file and returns a
    :class:`tuple <python:tuple>` with a Pandas
    :class:`DataFrame <pandas:pandas.DataFrame>` object and a metadata
    :class:`dict <python:dict>`.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete metadata
      :class:`dict <python:dict>`. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: A :class:`DataFrame <pandas:DataFrame>` representation of the SPSS data (or
      :obj:`None <python:None>`) and a :class:`Metadata` representation of the dataset's
      metadata / data map.
    :rtype: :class:`pandas.DataFrame <pandas:DataFrame>`/:obj:`None <python:None>` and
      :class:`Metadata`

    """
    if not any([checkers.is_file(data),
                checkers.is_bytesIO(data),
                checkers.is_type(data, bytes)]):
        raise errors.InvalidDataFormatError('data must be a filename, BytesIO, or bytes '
                                            f'object. Was: {data.__class__.__name__}')

    limit = validators.integer(limit, allow_empty = True, minimum = 0)
    offset = validators.integer(offset, minimum = 0)

    exclude_variables = validators.iterable(exclude_variables, allow_empty = True)
    if exclude_variables:
        exclude_variables = [validators.string(x) for x in exclude_variables]

    include_variables = validators.iterable(include_variables, allow_empty = True)
    if include_variables:
        include_variables = [validators.string(x) for x in include_variables]

    if not checkers.is_file(data):
        with tempfile.NamedTemporaryFile(delete = False) as temp_file:
            temp_file.write(data)
            temp_file_name = temp_file.name

        df, meta = pyreadstat.read_sav(temp_file_name,
                                       metadataonly = metadata_only,
                                       dates_as_pandas_datetime = dates_as_datetime64,
                                       apply_value_formats = apply_labels,
                                       formats_as_category = labels_as_categories,
                                       usecols = include_variables,
                                       user_missing = not missing_as_NaN,
                                       disable_datetime_conversion = not convert_datetimes,
                                       row_limit = limit or 0,
                                       row_offset = offset,
                                       **kwargs)
        os.remove(temp_file_name)
    else:
        df, meta = pyreadstat.read_sav(data,
                                       metadataonly = metadata_only,
                                       dates_as_pandas_datetime = dates_as_datetime64,
                                       apply_value_formats = apply_labels,
                                       formats_as_category = labels_as_categories,
                                       usecols = include_variables,
                                       user_missing = not missing_as_NaN,
                                       disable_datetime_conversion = not convert_datetimes,
                                       row_limit = limit or 0,
                                       row_offset = offset,
                                       **kwargs)

    metadata = Metadata.from_pyreadstat(meta)

    if exclude_variables:
        df = df.drop(exclude_variables, axis = 1)
        if metadata.column_metadata:
            for variable in exclude_variables:
                metadata.column_metadata.pop(variable, None)

    return df, metadata


def get_metadata(data):
    """Retrieve the metadata that describes the coded representation of the data,
    corresponding formatting information, and their related human-readable labels.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :returns: The metadata that describes the raw data and its corresponding labels.
    :rtype: :class:`Metadata`
    """
    return _read_spss(data, metadata_only = True)[1]


def to_dataframe(data: Union[bytes, BytesIO, 'os.PathLike[Any]'],
                 limit: Optional[int] = None,
                 offset: int = 0,
                 exclude_variables: Optional[List[str]] = None,
                 include_variables: Optional[List[str]] = None,
                 metadata_only: bool = False,
                 apply_labels: bool = False,
                 labels_as_categories: bool = True,
                 missing_as_NaN: bool = False,
                 convert_datetimes: bool = True,
                 dates_as_datetime64: bool = False,
                 **kwargs):
    """Reads SPSS data and returns a :class:`tuple <python:tuple>` with a Pandas
    :class:`DataFrame <pandas:pandas.DataFrame>` object and relevant :class:`Metadata`.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: A :class:`DataFrame <pandas:DataFrame>` representation of the SPSS data (or
      :obj:`None <python:None>`) and a :class:`Metadata <Metadata>` representation of the
      data's meta-data (value and labels / data map).
    :rtype: :class:`pandas.DataFrame <pandas:DataFrame>`/:obj:`None <python:None>` and
      :class:`Metadata <Metadata>`

    """
    return _read_spss(data,
                      limit = limit,
                      offset = offset,
                      exclude_variables = exclude_variables,
                      include_variables = include_variables,
                      metadata_only = metadata_only,
                      apply_labels = apply_labels,
                      labels_as_categories = labels_as_categories,
                      missing_as_NaN = missing_as_NaN,
                      convert_datetimes = convert_datetimes,
                      dates_as_datetime64 = dates_as_datetime64,
                      **kwargs)


def to_csv(data: Union['os.PathLike[Any]', BytesIO, bytes],
           target: Optional[Union['os.PathLike[Any]', StringIO]] = None,
           include_header: bool = True,
           delimter: str = '|',
           null_text: str = 'NaN',
           wrapper_character: str = "'",
           escape_character: str = "\\",
           line_terminator: str = '\r\n',
           decimal: str = '.',
           limit: Optional[int] = None,
           offset: int = 0,
           exclude_variables: Optional[List[str]] = None,
           include_variables: Optional[List[str]] = None,
           metadata_only: bool = False,
           apply_labels: bool = False,
           labels_as_categories: bool = True,
           missing_as_NaN: bool = False,
           convert_datetimes: bool = True,
           dates_as_datetime64: bool = False,
           **kwargs):
    r"""Convert the SPSS ``data`` into a CSV string where each row represents a record of
    SPSS data.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param target: The destination where the CSV representation should be stored. Accepts
      either a filename, file-pointer or a :class:`StringIO <python:io.StringIO>`, or
      :obj:`None <python:None>`. If :obj:`None <python:None>`, will return a
      :class:`str <python:str>` object stored in-memory. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`StringIO <python:io.StringIO>` /
      :class:`str <python:str>` / :obj:`None <python:None>`

    :param include_header: If ``True``, will include a header row with column
      labels. If ``False``, will not include a header row. Defaults to ``True``.
    :type include_header: :class:`bool <python:bool>`

    :param delimiter: The delimiter used between columns. Defaults to ``|``.
    :type delimiter: :class:`str <python:str>`

    :param null_text: The text value to use in place of empty values. Only
      applies if ``wrap_empty_values`` is ``True``. Defaults to ``'NaN'``.
    :type null_text: :class:`str <python:str>`

    :param wrapper_character: The string used to wrap string values when
      wrapping is necessary. Defaults to ``'``.
    :type wrapper_character: :class:`str <python:str>`

    :param escape_character: The character to use when escaping nested wrapper
      characters. Defaults to ``\``.
    :type escape_character: :class:`str <python:str>`

    :param line_terminator: The character used to mark the end of a line.
      Defaults to ``\r\n``.
    :type line_terminator: :class:`str <python:str>`

    :param decimal: The character used to indicate a decimal place in a numerical value.
      Defaults to ``.``.
    :type decimal: :class:`str <python:str>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: :obj:`None <python:None>` if ``target`` was not :obj:`None <python:None>`,
      otherwise a :class:`str <python:str>` representation of the CSV file.
    :rtype: :obj:`None <python:None>` or :class:`str <python:str>`

    """
    if target and not checkers.is_pathlike(target) and not checkers.is_stringIO(target):
        raise errors.InvalidDataFormatError('target must be a filename, StringIO object, '
                                            f'or None. Was: {data.__class__.__name__}')

    df, metadata = _read_spss(data,
                              limit = limit,
                              offset = offset,
                              exclude_variables = exclude_variables,
                              include_variables = include_variables,
                              metadata_only = metadata_only,
                              apply_labels = apply_labels,
                              labels_as_categories = labels_as_categories,
                              missing_as_NaN = missing_as_NaN,
                              convert_datetimes = convert_datetimes,
                              dates_as_datetime64 = dates_as_datetime64,
                              **kwargs)

    result = df.to_csv(target,
                       sep = delimter,
                       na_rep = null_text,
                       header = include_header,
                       quotechar = wrapper_character,
                       line_terminator = line_terminator,
                       escapechar = escape_character,
                       decimal = decimal)

    if target is not None:
        return None

    return result


def to_json(data: Union['os.PathLike[Any]', BytesIO, bytes],
            target: Optional[Union['os.PathLike[Any]', StringIO]] = None,
            layout: str = 'records',
            double_precision: int = 10,
            limit: Optional[int] = None,
            offset: int = 0,
            exclude_variables: Optional[List[str]] = None,
            include_variables: Optional[List[str]] = None,
            metadata_only: bool = False,
            apply_labels: bool = False,
            labels_as_categories: bool = True,
            missing_as_NaN: bool = False,
            convert_datetimes: bool = True,
            dates_as_datetime64: bool = False,
            **kwargs):
    r"""Convert the SPSS ``data`` into a JSON string.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param target: The destination where the JSON representation should be stored. Accepts
      either a filename, file-pointer or :class:`StringIO <python:io.StringIO>`, or
      :obj:`None <python:None>`. If :obj:`None <python:None>`, will return a
      :class:`str <python:str>` object stored in-memory. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`StringIO <python:io.StringIO>` /
      :class:`str <python:str>` / :obj:`None <python:None>`

    :param layout: Indicates the layout schema to use for the JSON representation of the
      data. Accepts:

        * ``records``, where the resulting JSON object represents an array of objects
          where each object corresponds to a single record, with key/value pairs for each
          column and that record's corresponding value
        * ``table``, where the resulting JSON object contains a metadata (data map)
          describing the data schema along with the resulting collection of record objects

      Defaults to ``records``.
    :type layout: :class:`str <python:str>`

    :param double_precision: Indicates the precision (places beyond the decimal point) to
      apply for floating point values. Defaults to ``10``.
    :type double_precision: class:`int <python:int>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: :obj:`None <python:None>` if ``target`` was not :obj:`None <python:None>`,
      otherwise a :class:`str <python:str>` representation of the JSON output.
    :rtype: :obj:`None <python:None>` or :class:`str <python:str>`

    """
    if target and not checkers.is_pathlike(target) and not checkers.is_stringIO(target):
        raise errors.InvalidDataFormatError('target must be a filename, StringIO object, '
                                            f'or None. Was: {data.__class__.__name__}')

    if layout not in ['records', 'table']:
        raise errors.InvalidLayoutError('layout must be either "records" or "table". '
                                        f'Was: "{layout}"')

    df, metadata = _read_spss(data,
                              limit = limit,
                              offset = offset,
                              exclude_variables = exclude_variables,
                              include_variables = include_variables,
                              metadata_only = metadata_only,
                              apply_labels = apply_labels,
                              labels_as_categories = labels_as_categories,
                              missing_as_NaN = missing_as_NaN,
                              convert_datetimes = convert_datetimes,
                              dates_as_datetime64 = dates_as_datetime64,
                              **kwargs)

    result = df.to_json(target,
                        orient = layout,
                        double_precision = double_precision)

    if target is not None:
        return None

    return result


def to_yaml(data: Union['os.PathLike[Any]', BytesIO, bytes],
            target: Optional[Union['os.PathLike[Any]', StringIO]] = None,
            layout: str = 'records',
            double_precision: int = 10,
            limit: Optional[int] = None,
            offset: int = 0,
            exclude_variables: Optional[List[str]] = None,
            include_variables: Optional[List[str]] = None,
            metadata_only: bool = False,
            apply_labels: bool = False,
            labels_as_categories: bool = True,
            missing_as_NaN: bool = False,
            convert_datetimes: bool = True,
            dates_as_datetime64: bool = False,
            **kwargs):
    r"""Convert the SPSS ``data`` into a YAML string.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param target: The destination where the YAML representation should be stored. Accepts
      either a filename, file-pointer or :class:`StringIO <python:io.StringIO>`, or
      :obj:`None <python:None>`. If :obj:`None <python:None>`, will return a
      :class:`str <python:str>` object stored in-memory. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`StringIO <python:io.StringIO>` /
      :class:`str <python:str>` / :obj:`None <python:None>`

    :param layout: Indicates the layout schema to use for the JSON representation of the
      data. Accepts:

        * ``records``, where the resulting YAML object represents an array of objects
          where each object corresponds to a single record, with key/value pairs for each
          column and that record's corresponding value
        * ``table``, where the resulting JSON object contains a metadata (data map)
          describing the data schema along with the resulting collection of record objects

      Defaults to ``records``.
    :type layout: :class:`str <python:str>`

    :param double_precision: Indicates the precision (places beyond the decimal point) to
      apply for floating point values. Defaults to ``10``.
    :type double_precision: class:`int <python:int>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: :obj:`None <python:None>` if ``target`` was not :obj:`None <python:None>`,
      otherwise a :class:`str <python:str>` representation of the YAML output.
    :rtype: :obj:`None <python:None>` or :class:`str <python:str>`

    """
    if target and not checkers.is_pathlike(target) and not checkers.is_stringIO(target):
        raise errors.InvalidDataFormatError('target must be a filename, StringIO object, '
                                            f'or None. Was: {data.__class__.__name__}')

    if layout not in ['records', 'table']:
        raise errors.InvalidLayoutError('layout must be either "records" or "table". '
                                        f'Was: "{layout}"')

    df, metadata = _read_spss(data,
                              limit = limit,
                              offset = offset,
                              exclude_variables = exclude_variables,
                              include_variables = include_variables,
                              metadata_only = metadata_only,
                              apply_labels = apply_labels,
                              labels_as_categories = labels_as_categories,
                              missing_as_NaN = missing_as_NaN,
                              convert_datetimes = convert_datetimes,
                              dates_as_datetime64 = dates_as_datetime64,
                              **kwargs)

    as_json = df.to_json(None,
                         orient = layout,
                         double_precision = double_precision)

    as_dict = json.loads(as_json)

    as_yaml = yaml.dump(as_dict)

    if target is None:
        return as_yaml

    with open(target, 'wb') as target_file:
        target_file.write(as_yaml)


def to_dict(data: Union['os.PathLike[Any]', BytesIO, bytes],
            layout: str = 'records',
            double_precision: int = 10,
            limit: Optional[int] = None,
            offset: int = 0,
            exclude_variables: Optional[List[str]] = None,
            include_variables: Optional[List[str]] = None,
            metadata_only: bool = False,
            apply_labels: bool = False,
            labels_as_categories: bool = True,
            missing_as_NaN: bool = False,
            convert_datetimes: bool = True,
            dates_as_datetime64: bool = False,
            **kwargs):
    r"""Convert the SPSS ``data`` into a Python :class:`dict <python:dict>`.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param layout: Indicates the layout schema to use for the JSON representation of the
      data. Accepts:

        * ``records``, where the resulting YAML object represents an array of objects
          where each object corresponds to a single record, with key/value pairs for each
          column and that record's corresponding value
        * ``table``, where the resulting JSON object contains a metadata (data map)
          describing the data schema along with the resulting collection of record objects

      Defaults to ``records``.
    :type layout: :class:`str <python:str>`

    :param double_precision: Indicates the precision (places beyond the decimal point) to
      apply for floating point values. Defaults to ``10``.
    :type double_precision: class:`int <python:int>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: :obj:`None <python:None>` if ``target`` was not :obj:`None <python:None>`,
      otherwise a :class:`list <python:list>` of :class:`dict <python:dict>` if ``layout``
      is ``records``, or a :class:`dict <python:dict>` if ``layout`` is ``table``.
    :rtype: :obj:`None <python:None>` or :class:`str <python:str>`

    """
    if layout not in ['records', 'table']:
        raise errors.InvalidLayoutError('layout must be either "records" or "table". '
                                        f'Was: "{layout}"')

    as_json = to_json(data,
                      layout = layout,
                      double_precision = double_precision,
                      limit = limit,
                      offset = offset,
                      exclude_variables = exclude_variables,
                      include_variables = include_variables,
                      metadata_only = metadata_only,
                      apply_labels = apply_labels,
                      labels_as_categories = labels_as_categories,
                      missing_as_NaN = missing_as_NaN,
                      convert_datetimes = convert_datetimes,
                      dates_as_datetime64 = dates_as_datetime64,
                      **kwargs)

    as_dict = json.loads(as_json)

    return as_dict


def to_excel(data: Union['os.PathLike[Any]', BytesIO, bytes],
             target: Optional[Union['os.PathLike[Any]', BytesIO, ExcelWriter]] = None,
             sheet_name: str = 'Sheet1',
             start_row: int = 0,
             start_column: int = 0,
             null_text: str = 'NaN',
             include_header: bool = True,
             limit: Optional[int] = None,
             offset: int = 0,
             exclude_variables: Optional[List[str]] = None,
             include_variables: Optional[List[str]] = None,
             metadata_only: bool = False,
             apply_labels: bool = False,
             labels_as_categories: bool = True,
             missing_as_NaN: bool = False,
             convert_datetimes: bool = True,
             dates_as_datetime64: bool = False,
             **kwargs):
    r"""Convert the SPSS ``data`` into an Excel file where each row represents a record of
    SPSS data.

    :param data: The SPSS data to load. Accepts either a series of bytes or a filename.
    :type data: Path-like filename, :class:`bytes <python:bytes>` or
      :class:`BytesIO <python:io.bytesIO>`

    :param target: The destination where the Excel file should be stored. Accepts
      either a filename, file-pointer or a :class:`BytesIO <python:io.BytesIO>`, or
      an :class:`ExcelWriter <pandas:pandas.ExcelWriter>` instance.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :class:`ExcelWriter <pandas:pandas.ExcelWriter>`

    :param sheet_name: The worksheet on which the SPSS data should be written. Defaults to
      ``'Sheet1'``.
    :type sheet_name: :class:`str <python:str>`

    :param start_row: The row number (starting at 0) where the SPSS data should begin.
      Defaults to ``0``.
    :type start_row: :class:`int <python:int>`

    :param start_column: The column number (starting at 0) where the SPSS data should
      begin. Defaults to ``0``.
    :type start_column: :class:`int <python:int>`

    :param null_text: The way that missing values should be represented in the Excel
      file. Defaults to ``''`` (an empty string).
    :type null_text: :class:`str <python:str>`

    :param include_header: If ``True``, will include a header row with column
      labels. If ``False``, will not include a header row. Defaults to ``True``.
    :type include_header: :class:`bool <python:bool>`

    :param limit: The number of records to read from the data. If
      :obj:`None <python:None>` will return all records. Defaults to
      :obj:`None <python:None>`.
    :type limit: :class:`int <python:int>` or :obj:`None <python:None>`

    :param offset: The record at which to start reading the data. Defaults to 0 (first
      record).
    :type offset: :class:`int <python:int>`

    :param exclude_variables: A list of the variables that should be ignored when reading
      data. Defaults to :obj:`None <python:None>`.
    :type exclude_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param include_variables: A list of the variables that should be explicitly included
      when reading data. Defaults to :obj:`None <python:None>`.
    :type include_variables: iterable of :class:`str <python:str>` or
      :obj:`None <python:None>`

    :param metadata_only: If ``True``, will return no data records in the resulting
      :class:`DataFrame <pandas:pandas.DataFrame>` but will return a complete
      :class:`Metadata` instance. Defaults to ``False``.
    :type metadata_only: :class:`bool <python:bool>`

    :param apply_labels: If ``True``, converts the numerically-coded values in the raw
      data to their human-readable labels. Defaults to ``False``.
    :type apply_labels: :class:`bool <python:bool>`

    :param labels_as_categories: If ``True``, will convert labeled or formatted values to
      Pandas :term:`categories <pandas:category>`. Defaults to ``True``.

      .. caution::

        This parameter will only have an effect if the ``apply_labels`` parameter is
        ``True``.

    :type labels_as_categories: :class:`bool <python:bool>`

    :param missing_as_NaN: If ``True``, will return any missing values as
      :class:`NaN <pandas:NaN>`. Otherwise will return missing values as per the
      configuration of missing value representation stored in the underlying SPSS data.
      Defaults to ``False``, which applies the missing value representation configured in
      the SPSS data itself.
    :type missing_as_NaN: :class:`bool <python:bool>`

    :param convert_datetimes: if ``True``, will convert the native integer representation
      of datetime values in the SPSS data to Pythonic
      :class:`datetime <python:datetime.datetime>`, or
      :class:`date <python:datetime.date>`, etc. representations (or Pandas
      :class:`datetime64 <pandas:datetime64>`, depending on the ``dates_as_datetime64``
      parameter). If ``False``, will leave the original integer representation. Defaults
      to ``True``.
    :type convert_datetimes: :class:`bool <python:bool>`

    :param dates_as_datetime64: If ``True``, will return any date values as Pandas
      :class:`datetime64 <pandas.datetime64>` types. Defaults to ``False``.

      .. caution::

        This parameter is only applied if ``convert_datetimes`` is set to ``True``.

    :type dates_as_datetime64: :class:`bool <python:bool>`

    :returns: :obj:`None <python:None>` if ``target`` was not :obj:`None <python:None>`,
      otherwise a :class:`BytesIO <python:BytesIO>` representation of the Excel file.
    :rtype: :obj:`None <python:None>` or :class:`str <python:str>`

    """
    if target and \
       not checkers.is_pathlike(target) and \
       not checkers.is_bytesIO(target) and \
       not checkers.is_type(target, 'ExcelWriter'):
        raise errors.InvalidDataFormatError('target must be a filename, BytesIO, '
                                            f'ExcelWriter, or None. '
                                            f'Was: {data.__class__.__name__}')

    df, metadata = _read_spss(data,
                              limit = limit,
                              offset = offset,
                              exclude_variables = exclude_variables,
                              include_variables = include_variables,
                              metadata_only = metadata_only,
                              apply_labels = apply_labels,
                              labels_as_categories = labels_as_categories,
                              missing_as_NaN = missing_as_NaN,
                              convert_datetimes = convert_datetimes,
                              dates_as_datetime64 = dates_as_datetime64,
                              **kwargs)

    return_target = False
    if not target or checkers.is_bytesIO(target):
        return_target = True
        target = BytesIO()

    df.to_excel(target,
                sheet_name = sheet_name,
                na_rep = null_text,
                header = include_header,
                startrow = start_row,
                startcol = start_column)

    if return_target:
        return target



from typing import Union, Optional, Any
from os import PathLike
import tempfile

from io import BytesIO
import yaml
import simplejson as json
import pyreadstat
from validator_collection import validators, checkers
from spss_converter.Metadata import Metadata
import pandas

DataFrame = pandas.DataFrame


def from_dataframe(df: DataFrame,
                   target: Optional[Union['PathLike[Any]', BytesIO]] = None,
                   metadata: Optional[Metadata] = None,
                   compress: bool = False):
    """Create an SPSS dataset from a `Pandas <https://pandas.pydata.org/>`_
    :class:`DataFrame <pandas:DataFrame>`.

    :param df: The :class:`DataFrame` to serialize to an SPSS dataset.
    :type df: :class:`pandas.DataFrame <pandas:DataFrame>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param metadata: The :class:`Metadata` associated with the dataset. If
      :obj:`None <python:None>`, will attempt to derive it form ``df``. Defaults to
      :obj:`None <python:None>`.
    :type metadata: :class:`Metadata` / :obj:`None <python:None>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    :raises ValueError: if ``df`` is not a :class:`pandas.DataFrame <pandas:DataFrame>`
    :raises ValueError: if ``metadata`` is not a :class:`Metadata`

    """
    if not checkers.is_type(df, 'DataFrame'):
        raise ValueError(f'df must be a pandas.DataFrame. Was: {df.__class__.__name__}')
    if metadata and not checkers.is_type(metadata, ('Metadata',
                                                    'metadata_container',
                                                    'dict')):
        raise ValueError(f'metadata must be a Metadata instance or compatible object. '
                         f'Was: {metadata.__class__.__name__}')
    elif metadata and checkers.is_type(metadata, 'metadata_container'):
        metadata = Metadata.from_pyreadstat(metadata)
    elif metadata and checkers.is_type(metadata, 'dict'):
        metadata = Metadata.from_dict(metadata)

    is_file = False
    if target and checkers.is_pathlike(target):
        is_file = True
    elif target:
        target = validators.bytesIO(target, allow_empty = False)

    if metadata:
        as_pyreadstat = metadata.to_pyreadstat()
    else:
        as_pyreadstat = None

    if target and is_file:
        if as_pyreadstat:
            pyreadstat.write_sav(
                df = df,
                dst_path = target,
                file_label = as_pyreadstat.file_label,
                column_labels = as_pyreadstat.column_labels,
                compress = compress,
                note = as_pyreadstat.notes,
                variable_value_labels = as_pyreadstat.variable_value_labels,
                missing_ranges = as_pyreadstat.missing_ranges,
                variable_display_width = as_pyreadstat.variable_display_width,
                variable_measure = as_pyreadstat.variable_measure
            )
        else:
            pyreadstat.write_sav(df = df,
                                 dst_path = target,
                                 compress = compress)

    else:
        with tempfile.NamedTemporaryFile() as temp_file:
            if as_pyreadstat:
                pyreadstat.write_sav(
                    df = df,
                    dst_path = temp_file.name,
                    file_label = as_pyreadstat.file_label,
                    column_labels = as_pyreadstat.column_labels,
                    compress = compress,
                    note = as_pyreadstat.notes,
                    variable_value_labels = as_pyreadstat.variable_value_labels,
                    missing_ranges = as_pyreadstat.missing_ranges,
                    variable_display_width = as_pyreadstat.variable_display_width,
                    variable_measure = as_pyreadstat.variable_measure
                )
            else:
                pyreadstat.write_sav(df = df,
                                     dst_path = temp_file.name,
                                     compress = compress)

            if target:
                target.write(temp_file.read())
            else:
                target = BytesIO(temp_file.read())

            return target


def from_csv(as_csv: Union[str, 'PathLike[Any]', BytesIO],
             target: Optional[Union['PathLike[Any]', BytesIO]] = None,
             compress: bool = False,
             delimiter = '|',
             **kwargs):
    """Convert a CSV file into an SPSS dataset.

    .. tip::

      If you pass any additional keyword arguments, those keyword arguments will be passed
      onto the :func:`pandas.read_csv() <pandas:pandas.read_csv>` function.

    :param as_csv: The CSV data that you wish to convert into an SPSS dataset.
    :type as_csv: :class:`str <python:str>` / File-location /
      :class:`BytesIO <python:io.BytesIO>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :param delimiter: The delimiter used between columns. Defaults to ``|``.
    :type delimiter: :class:`str <python:str>`

    :param kwargs: Additional keyword arguments which will be passed onto the
      :func:`pandas.read_csv() <pandas:pandas.read_csv>` function.
    :type kwargs: :class:`dict <python:dict>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    """
    df = pandas.read_csv(as_csv,
                         delimiter = delimiter,
                         **kwargs)
    if 'Unnamed: 0' in df:
        df = df.drop(['Unnamed: 0'], axis = 1)

    result = from_dataframe(df,
                            target = target,
                            compress = compress)

    return result


def from_dict(as_dict: dict,
              target: Optional[Union['PathLike[Any]', BytesIO]] = None,
              compress: bool = False,
              **kwargs):
    """Convert a :class:`dict <python:dict>` object into an SPSS dataset.

    .. tip::

      If you pass any additional keyword arguments, those keyword arguments will be passed
      onto the :meth:`DataFrame.from_dict() <pandas:pandas.DataFrame.from_dict>` method.

    :param as_dict: The :class:`dict <python:dict>` data that you wish to convert into an
      SPSS dataset.
    :type as_dict: :class:`dict <python:dict>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :param kwargs: Additional keyword arguments which will be passed onto the
      :meth:`DataFrame.from_dict() <pandas:pandas.DataFrame.from_dict>` method.
    :type kwargs: :class:`dict <python:dict>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    """
    df = DataFrame.from_dict(as_dict, **kwargs)

    result = from_dataframe(df,
                            target = target,
                            compress = compress)

    return result


def from_json(as_json: Union[str, 'PathLike[Any]', BytesIO],
              target: Optional[Union['PathLike[Any]', BytesIO]] = None,
              layout: str = 'records',
              compress: bool = False,
              **kwargs):
    """Convert JSON data into an SPSS dataset.

    .. tip::

      If you pass any additional keyword arguments, those keyword arguments will be passed
      onto the :func:`pandas.read_json() <pandas:pandas.read_json>` function.

    :param as_json: The JSON data that you wish to convert into an SPSS dataset.
    :type as_json: :class:`str <python:str>` / File-location /
      :class:`BytesIO <python:io.BytesIO>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param layout: Indicates the layout schema to use for the JSON representation of the
      data. Accepts:

        * ``records``, where the resulting JSON object represents an array of objects
          where each object corresponds to a single record, with key/value pairs for each
          column and that record's corresponding value
        * ``table``, where the resulting JSON object contains a metadata (data map)
          describing the data schema along with the resulting collection of record objects

      Defaults to ``records``.

    :type layout: :class:`str <python:str>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :param kwargs: Additional keyword arguments which will be passed onto the
      :func:`pandas.read_json() <pandas:pandas.read_json>` function.

      .. warning::

        If you supply an ``orient`` keyword argument (which is supported by
        :func:`pandas.read_json() <pandas:pandas.read_json>`), the ``orient``
        value will *override* the value supplied for ``layout``. This is an
        advanced use case, so use with caution.

    :type kwargs: :class:`dict <python:dict>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    """
    if layout not in ['records', 'table']:
        raise errors.InvalidLayoutError('layout must be either "records" or "table". '
                                        f'Was: "{layout}"')

    orient = kwargs.pop('orient', layout)

    df = pandas.read_json(as_json,
                          orient = orient,
                          **kwargs)

    result = from_dataframe(df,
                            target = target,
                            compress = compress)

    return result


def from_yaml(as_yaml: Union[str, 'PathLike[Any]', BytesIO],
              target: Optional[Union['PathLike[Any]', BytesIO]] = None,
              layout: str = 'records',
              compress: bool = False,
              **kwargs):
    """Convert YAML data into an SPSS dataset.

    .. tip::

      If you pass any additional keyword arguments, those keyword arguments will be passed
      onto the :meth:`DataFrame.from_dict() <pandas:pandas.DataFrame.from_dict>` method.

    :param as_yaml: The YAML data that you wish to convert into an SPSS dataset.
    :type as_yaml: :class:`str <python:str>` / File-location /
      :class:`BytesIO <python:io.BytesIO>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param layout: Indicates the layout schema to expect for the YAML representation of the
      data. Accepts:

        * ``records``, where the resulting JSON object represents an array of objects
          where each object corresponds to a single record, with key/value pairs for each
          column and that record's corresponding value
        * ``table``, where the resulting JSON object contains a metadata (data map)
          describing the data schema along with the resulting collection of record objects

      Defaults to ``records``.
    :type layout: :class:`str <python:str>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :param kwargs: Additional keyword arguments which will be passed onto the
      :meth:`pandas.from_json() <pandas:pandas.from_json>` function.

    :param kwargs: Additional keyword arguments which will be passed onto the
      :func:`pandas.read_json() <pandas:pandas.read_json>` function.

      .. warning::

        If you supply an ``orient`` keyword argument (which is supported by
        :func:`pandas.read_json() <pandas:pandas.read_json>`), the ``orient``
        value will *override* the value supplied for ``layout``. This is an
        advanced use case, so use with caution.

    :type kwargs: :class:`dict <python:dict>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    """
    if checkers.is_file(as_yaml) or checkers.is_bytesIO(as_yaml):
        file_path = as_yaml
        with open(file_path, 'rb') as yaml_file:
            as_dict = yaml.safe_load(yaml_file)
    else:
        as_yaml = validators.string(as_yaml, allow_empty = False)
        as_dict = yaml.safe_load(as_yaml)
        as_json = json.dumps(as_dict)

    return from_json(as_json,
                     target = target,
                     layout = layout,
                     compress = compress,
                     **kwargs)


def from_excel(as_excel,
               target: Optional[Union['PathLike[Any]', BytesIO]] = None,
               compress: bool = False,
               **kwargs):
    """Convert Excel data into an SPSS dataset.

    .. tip::

      If you pass any additional keyword arguments, those keyword arguments will be passed
      onto the :func:`pandas.read_excel() <pandas:pandas.read_excel>` function.

    :param as_excel: The Excel data that you wish to convert into an SPSS dataset.
    :type as_excel: :class:`str <python:str>` / File-location /
      :class:`BytesIO <python:io.BytesIO>` / :class:`bytes <python:bytes>` /
      :class:`ExcelFile <pandas.ExcelFile>`

    :param target: The target to which the SPSS dataset should be written. Accepts either
      a filename/path, a :class:`BytesIO <python:io.BytesIO>` object, or
      :obj:`None <python:None>`. If :obj:`None <python:None>` will return a
      :class:`BytesIO <python:io.BytesIO>` object containing the SPSS dataset. Defaults to
      :obj:`None <python:None>`.
    :type target: Path-like / :class:`BytesIO <python:io.BytesIO>` /
      :obj:`None <python:None>`

    :param compress: If ``True``, will return data in the compressed ZSAV format. If
      ``False``, will return data in the standards SAV format. Defaults to ``False``.
    :type compress: :class:`bool <python:bool>`

    :param kwargs: Additional keyword arguments which will be passed onto the
      :func:`pandas.read_excel() <pandas:pandas.read_excel>` function.
    :type kwargs: :class:`dict <python:dict>`

    :returns: A :class:`BytesIO <python:io.BytesIO>` object containing the SPSS data if
      ``target`` is :obj:`None <python:None>` or not a filename, otherwise
      :obj:`None <python:None>`
    :rtype: :class:`BytesIO <python:io.BytesIO>` or :obj:`None <python:None>`

    """
    df = pandas.read_excel(as_excel, **kwargs)
    if 'Unnamed: 0' in df:
        df = df.drop(['Unnamed: 0'], axis = 1)

    result = from_dataframe(df,
                            target = target,
                            compress = compress)

    return result


def apply_metadata(df: DataFrame,
                   metadata: Union[Metadata, dict, pyreadstat.metadata_container],
                   as_category: bool = True):
    """Updates the :class:`DataFrame <pandas:DataFrame>` ``df`` based on the ``metadata``.

    :param df: The :class:`DataFrame <pandas:pandas.DataFrame>` to update.
    :type df: :class:`pandas.DataFrame <pandas:pandas.DataFrame>`

    :param metadata: The :class:`Metadata` to apply to ``df``.
    :type metadata: :class:`Metadata`, :class:`pyreadstat.metadata_container`, or
      compatible :class:`dict <python:dict>`

    :param as_category: if ``True``, will variables with formats will be transformed into
      categories in the :class:`DataFrame <pandas:pandas.DataFrame>`. Defaults to
      ``True``.
    :type as_category: :class:`bool <python:bool>`

    :returns: A copy of ``df`` updated to reflect ``metadata``.
    :rtype: :class:`DataFrame <pandas:pandas.DataFrame>`
    """
    if not checkers.is_type(df, 'DataFrame'):
        raise ValueError(f'df must be a pandas.DataFrame. Was: {df.__class__.__name__}')
    if not checkers.is_type(metadata, ('Metadata', 'metadata_container', 'dict')):
        raise ValueError(f'metadata must be a Metadata instance or compatible object. '
                         f'Was: {metadata.__class__.__name__}')
    elif checkers.is_type(metadata, 'metadata_container'):
        metadata = Metadata.from_pyreadstat(metadata)
    elif checkers.is_type(metadata, 'dict'):
        metadata = Metadata.from_dict(metadata)

    as_pyreadstat = metadata.to_pyreadstat()

    return pyreadstat.set_value_labels(df,
                                       metadata = as_pyreadstat,
                                       formats_as_category = as_category)
